@@ -5,8 +5,7 @@ use std::fmt::{self, Debug, Display};
 use bitflags::bitflags;
 use glutin::event::VirtualKeyCode::*;
 use glutin::event::{ModifiersState, MouseButton, VirtualKeyCode};
-use serde::de::Error as SerdeError;
-use serde::de::{self, MapAccess, Unexpected, Visitor};
+use serde::de::{self, Error as SerdeError, MapAccess, Unexpected, Visitor};
 use serde::{Deserialize, Deserializer};
 use serde_yaml::Value as SerdeValue;
 
@@ -109,6 +108,10 @@ pub enum Action {
     #[config(skip)]
     Search(SearchAction),
 
+    /// Perform mouse binding exclusive action.
+    #[config(skip)]
+    Mouse(MouseAction),
+
     /// Paste contents of system clipboard.
     Paste,
 
@@ -177,6 +180,9 @@ pub enum Action {
     /// Spawn a new instance of Alacritty.
     SpawnNewInstance,
 
+    /// Create a new Alacritty window.
+    CreateNewWindow,
+
     /// Toggle fullscreen.
     ToggleFullscreen,
 
@@ -227,12 +233,19 @@ impl From<SearchAction> for Action {
     }
 }
 
+impl From<MouseAction> for Action {
+    fn from(action: MouseAction) -> Self {
+        Self::Mouse(action)
+    }
+}
+
 /// Display trait used for error logging.
 impl Display for Action {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         match self {
             Action::ViMotion(motion) => motion.fmt(f),
             Action::Vi(action) => action.fmt(f),
+            Action::Mouse(action) => action.fmt(f),
             _ => write!(f, "{:?}", self),
         }
     }
@@ -283,6 +296,13 @@ pub enum SearchAction {
     SearchHistoryNext,
 }
 
+/// Mouse binding specific actions.
+#[derive(ConfigDeserialize, Debug, Copy, Clone, PartialEq, Eq)]
+pub enum MouseAction {
+    /// Expand the selection to the current mouse cursor position.
+    ExpandSelection,
+}
+
 macro_rules! bindings {
     (
         KeyBinding;
@@ -343,6 +363,8 @@ macro_rules! bindings {
 pub fn default_mouse_bindings() -> Vec<MouseBinding> {
     bindings!(
         MouseBinding;
+        MouseButton::Right;                         MouseAction::ExpandSelection;
+        MouseButton::Right,   ModifiersState::CTRL; MouseAction::ExpandSelection;
         MouseButton::Middle, ~BindingMode::VI; Action::PasteSelection;
     )
 }
@@ -1027,6 +1049,9 @@ impl<'a> Deserialize<'a> for RawBinding {
                                 SearchAction::deserialize(value.clone())
                             {
                                 Some(search_action.into())
+                            } else if let Ok(mouse_action) = MouseAction::deserialize(value.clone())
+                            {
+                                Some(mouse_action.into())
                             } else {
                                 match Action::deserialize(value.clone()).map_err(V::Error::custom) {
                                     Ok(action) => Some(action),
@@ -1078,7 +1103,7 @@ impl<'a> Deserialize<'a> for RawBinding {
 
                 let mode = mode.unwrap_or_else(BindingMode::empty);
                 let not_mode = not_mode.unwrap_or_else(BindingMode::empty);
-                let mods = mods.unwrap_or_else(ModifiersState::default);
+                let mods = mods.unwrap_or_default();
 
                 let action = match (action, chars, command) {
                     (Some(action @ Action::ViMotion(_)), None, None)
@@ -1102,13 +1127,22 @@ impl<'a> Deserialize<'a> for RawBinding {
                         }
                         action
                     },
+                    (Some(action @ Action::Mouse(_)), None, None) => {
+                        if mouse.is_none() {
+                            return Err(V::Error::custom(format!(
+                                "action `{}` is only available for mouse bindings",
+                                action,
+                            )));
+                        }
+                        action
+                    },
                     (Some(action), None, None) => action,
                     (None, Some(chars), None) => Action::Esc(chars),
                     (None, None, Some(cmd)) => Action::Command(cmd),
                     _ => {
                         return Err(V::Error::custom(
                             "must specify exactly one of chars, action or command",
-                        ))
+                        ));
                     },
                 };
 
