@@ -141,11 +141,6 @@ pub struct MessageBuffer {
 }
 
 impl MessageBuffer {
-    /// Create new message buffer.
-    pub fn new() -> MessageBuffer {
-        MessageBuffer { messages: VecDeque::new() }
-    }
-
     /// Check if there are any messages queued.
     #[inline]
     pub fn is_empty(&self) -> bool {
@@ -196,7 +191,7 @@ mod tests {
     #[test]
     fn appends_close_button() {
         let input = "a";
-        let mut message_buffer = MessageBuffer::new();
+        let mut message_buffer = MessageBuffer::default();
         message_buffer.push(Message::new(input.into(), MessageType::Error));
         let size = SizeInfo::new(7., 10., 1., 1., 0., 0., false);
 
@@ -208,7 +203,7 @@ mod tests {
     #[test]
     fn multiline_close_button_first_line() {
         let input = "fo\nbar";
-        let mut message_buffer = MessageBuffer::new();
+        let mut message_buffer = MessageBuffer::default();
         message_buffer.push(Message::new(input.into(), MessageType::Error));
         let size = SizeInfo::new(6., 10., 1., 1., 0., 0., false);
 
@@ -220,7 +215,7 @@ mod tests {
     #[test]
     fn splits_on_newline() {
         let input = "a\nb";
-        let mut message_buffer = MessageBuffer::new();
+        let mut message_buffer = MessageBuffer::default();
         message_buffer.push(Message::new(input.into(), MessageType::Error));
         let size = SizeInfo::new(6., 10., 1., 1., 0., 0., false);
 
@@ -232,7 +227,7 @@ mod tests {
     #[test]
     fn splits_on_length() {
         let input = "foobar1";
-        let mut message_buffer = MessageBuffer::new();
+        let mut message_buffer = MessageBuffer::default();
         message_buffer.push(Message::new(input.into(), MessageType::Error));
         let size = SizeInfo::new(6., 10., 1., 1., 0., 0., false);
 
@@ -244,7 +239,7 @@ mod tests {
     #[test]
     fn empty_with_shortterm() {
         let input = "foobar";
-        let mut message_buffer = MessageBuffer::new();
+        let mut message_buffer = MessageBuffer::default();
         message_buffer.push(Message::new(input.into(), MessageType::Error));
         let size = SizeInfo::new(6., 0., 1., 1., 0., 0., false);
 
@@ -256,7 +251,7 @@ mod tests {
     #[test]
     fn truncates_long_messages() {
         let input = "hahahahahahahahahahaha truncate this because it's too long for the term";
-        let mut message_buffer = MessageBuffer::new();
+        let mut message_buffer = MessageBuffer::default();
         message_buffer.push(Message::new(input.into(), MessageType::Error));
         let size = SizeInfo::new(22., (MIN_FREE_LINES + 2) as f32, 1., 1., 0., 0., false);
 
@@ -271,7 +266,7 @@ mod tests {
     #[test]
     fn hide_button_when_too_narrow() {
         let input = "ha";
-        let mut message_buffer = MessageBuffer::new();
+        let mut message_buffer = MessageBuffer::default();
         message_buffer.push(Message::new(input.into(), MessageType::Error));
         let size = SizeInfo::new(2., 10., 1., 1., 0., 0., false);
 
@@ -283,7 +278,7 @@ mod tests {
     #[test]
     fn hide_truncated_when_too_narrow() {
         let input = "hahahahahahahahaha";
-        let mut message_buffer = MessageBuffer::new();
+        let mut message_buffer = MessageBuffer::default();
         message_buffer.push(Message::new(input.into(), MessageType::Error));
         let size = SizeInfo::new(2., (MIN_FREE_LINES + 2) as f32, 1., 1., 0., 0., false);
 
@@ -295,7 +290,7 @@ mod tests {
     #[test]
     fn add_newline_for_button() {
         let input = "test";
-        let mut message_buffer = MessageBuffer::new();
+        let mut message_buffer = MessageBuffer::default();
         message_buffer.push(Message::new(input.into(), MessageType::Error));
         let size = SizeInfo::new(5., 10., 1., 1., 0., 0., false);
 
@@ -306,7 +301,7 @@ mod tests {
 
     #[test]
     fn remove_target() {
-        let mut message_buffer = MessageBuffer::new();
+        let mut message_buffer = MessageBuffer::default();
         for i in 0..10 {
             let mut msg = Message::new(i.to_string(), MessageType::Error);
             if i % 2 == 0 && i < 5 {
@@ -329,7 +324,7 @@ mod tests {
 
     #[test]
     fn pop() {
-        let mut message_buffer = MessageBuffer::new();
+        let mut message_buffer = MessageBuffer::default();
         let one = Message::new(String::from("one"), MessageType::Error);
         message_buffer.push(one.clone());
         let two = Message::new(String::from("two"), MessageType::Warning);
@@ -345,7 +340,7 @@ mod tests {
     #[test]
     fn wrap_on_words() {
         let input = "a\nbc defg";
-        let mut message_buffer = MessageBuffer::new();
+        let mut message_buffer = MessageBuffer::default();
         message_buffer.push(Message::new(input.into(), MessageType::Error));
         let size = SizeInfo::new(5., 10., 1., 1., 0., 0., false);
 
@@ -361,7 +356,7 @@ mod tests {
     #[test]
     fn wrap_with_unicode() {
         let input = "ab\nc 👩d fgh";
-        let mut message_buffer = MessageBuffer::new();
+        let mut message_buffer = MessageBuffer::default();
         message_buffer.push(Message::new(input.into(), MessageType::Error));
         let size = SizeInfo::new(7., 10., 1., 1., 0., 0., false);
 
@@ -377,7 +372,7 @@ mod tests {
     #[test]
     fn strip_whitespace_at_linebreak() {
         let input = "\n0 1 2 3";
-        let mut message_buffer = MessageBuffer::new();
+        let mut message_buffer = MessageBuffer::default();
         message_buffer.push(Message::new(input.into(), MessageType::Error));
         let size = SizeInfo::new(3., 10., 1., 1., 0., 0., false);
 
@@ -388,7 +383,7 @@ mod tests {
 
     #[test]
     fn remove_duplicates() {
-        let mut message_buffer = MessageBuffer::new();
+        let mut message_buffer = MessageBuffer::default();
         for _ in 0..10 {
             let msg = Message::new(String::from("test"), MessageType::Error);
             message_buffer.push(msg);
